/*
 * $Id: cmdline.c,v 1.3.2.2 1997/11/13 20:43:44 randall Exp $
 */

/*
 * $Log: cmdline.c,v $
 * Revision 1.3.2.2  1997/11/13 20:43:44  randall
 * Debugging support enhanced.
 *
 * Revision 1.3.2.1  1997/06/13 22:33:19  randall
 * Added distributed version to the repository under
 * branch "distributed"
 *
 * Revision 1.3  1997/06/09 19:03:11  athena
 * Changed RTS so that Cilk won't dump core by default
 *
 * Revision 1.2  1997/03/01 17:42:49  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/***********************************************************\
 * Command line parsing
\***********************************************************/

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <stdio.h>
#include <ctype.h>

enum {
     NONE, NPROC, STATS, NO_STATS, HELP, ARENA, CACHE, CACHEDIR, BSTORE, STACK, YIELD, NO_YIELD,
     DUMP_CORE, NO_DUMP_CORE, END_OPTIONS
};

static struct options {
     char *string;
     int option;
     char *help;
} optarray[] = {
     {
	  "-", END_OPTIONS, "-- : end of option parsing"
     },
     {
          "nproc", NPROC, "-nproc [<machine>] [<n>] : use <n> processes on machine <machine>.  <n> defaults to 1 and <machine> defaults to the local machine"
     },
     {
          "proc", NPROC, "-proc: same as -nproc"
     },
     {
	  "stack", STACK, "-stack <n> : set stack depth"
     },
     {
	  "stats", STATS, "-stats <n> : set statistics level"
     },
     {
	  "no-stats", NO_STATS, "-no-stats : disable statistics printing"
     },
     {
	  "help", HELP, "-help : print this message"
     },
     {
	  "heap", ARENA, "-heap <n> : set heap size (in MB)"
     },
     {
	  "cache", CACHE, "-cache <n> : set cache size per machine (in MB)"
     },
     {
          "cachedir", CACHEDIR, "-cachedir <dir> : set directory in which cache file is allocated"
     },
     {
          "bstore", BSTORE, "-bstore <n> : set backing store size (in MB)"
     },
     {
	  "yield", YIELD, "-yield : yield time slice when steal fails"
     },
     {
	  "no-yield", NO_YIELD, "-no-yield : don't yield time slice"
     },
     {
	  "dump-core", DUMP_CORE, "-dump-core : dump core on fatal cilk errors"
     },
     {
	  "no-dump-core", NO_DUMP_CORE, 
	       "-no-dump-core : do not dump core on fatal cilk errors"
     },
     {
	  (char *) 0, NONE, ""
     }
};

static void print_help(void)
{
     struct options *p;

     fprintf(stderr, "Cilk options:\n");

     for (p = optarray; p->string; ++p)
	  if (p->help)
	       fprintf(stderr, "     %s\n", p->help);

     fprintf(stderr, "\n");
}

/* look for a given string in the option table */
static struct options *parse_option(char *s)
{
     struct options *p;

     for (p = optarray; p->string; ++p)
	  if (strcmp(s, p->string) == 0)
	       break;

     return p;
}

#define CHECK(cond, complaint) \
     if (!(cond)) { fprintf(stderr, "Bad option argument for -%s: %s\n", p->string, complaint); return 1; }

char *Cilk_option_machine[MAXMACHINES];
int Cilk_option_nproc[MAXMACHINES];

int Cilk_parse_command_line(Options *options, int argc, char *argv[])
{
     int i, j;
     struct options *p;

     /* default options */
     *options = DEFAULT_OPTIONS;

     for (i = 1; i < argc; ++i) {
	  if (*argv[i] != '-')
	       goto arch_specific;
	  p = parse_option(argv[i] + 1);
	  switch (p->option) {
	      case NPROC:
		   CILK_ASSERT(options->nmach < MAXMACHINES);
		   ++i;
		   CHECK(i < argc,
			 "argument missing");
		   if (isalpha(argv[i][0]))
		     {
		       /* next argument is a machine name */
		       Cilk_option_machine[options->nmach] = argv[i];
		       
		       ++i;
		       if (i >= argc || !isdigit(argv[i][0]))
			 {
			   Cilk_option_nproc[options->nmach] = DEFAULT_NPROC;
			   --i;
			 }
		       else
			 {
			   j = atoi(argv[i]);
			   if (j == 0)
			     j = Cilk_partition_size();
			   CHECK(j > 0,
				 "non-positive number of machines");
			   /* note: +1 for communication processor */
			   Cilk_option_nproc[options->nmach] = j;
			 }
		     }
		   else
		     {
		       CHECK(isdigit(argv[i][0]),
			     "bad argument");
		       j = atoi(argv[i]);
		       if (j == 0)
			 j = Cilk_partition_size();
		       CHECK(j > 0,
			     "non-positive number of machines");
		       
		       Cilk_option_machine[options->nmach] = NULL;
		       Cilk_option_nproc[options->nmach] = j;
		     }
		   options->nmach++;
		   break;
	      case STACK:
		   ++i;
		   CHECK(i < argc,
			 "argument missing");
		   options->stackdepth = atoi(argv[i]);
		   CHECK(options->stackdepth > 0,
			 "non-positive stack depth");
		   break;
	      case STATS:
		   ++i;
		   CHECK(i < argc,
			 "argument missing");
		   options->statlevel = atoi(argv[i]);
		   break;
	      case NO_STATS:
		   options->statlevel = 0;
		   break;
	      case YIELD:
		   options->yieldslice = 1;
		   break;
	      case NO_YIELD:
		   options->yieldslice = 0;
		   break;
	      case DUMP_CORE:
		   options->dump_core = 1;
		   break;
	      case NO_DUMP_CORE:
		   options->dump_core = 0;
		   break;
	      case HELP:
		   print_help();
		   return 1;
		   break;
	      case ARENA:
		   ++i;
		   CHECK(i < argc,
			 "argument missing");
		   options->heap = atof(argv[i]) * MEGABYTE;
		   CHECK(options->heap > 0,
			 "non-positive heap size");
		   break;
	      case CACHE:
		   ++i;
		   CHECK(i < argc,
			 "argument missing");
		   options->cache = atof(argv[i]) * MEGABYTE;
		   CHECK(options->cache > 0,
			 "non-positive cache size");
		   break;
	      case CACHEDIR:
		   ++i;
		   CHECK(i < argc,
			 "argument missing");
		   strcpy(options->cachedir, argv[i]);
		   break;
	      case BSTORE:
		   ++i;
		   CHECK(i < argc,
			 "argument missing");
		   options->bstore = atof(argv[i]) * MEGABYTE;
		   CHECK(options->bstore > 0,
			 "non-positive bstore size");
		   break;
	      case END_OPTIONS:
		   ++i;
		   goto stop;
	      default:
		   goto arch_specific;
	  }
     }

     /* let architecture-specific code get a crack at the command line */
   arch_specific:
     i = Cilk_arch_specific_options(i, argc, argv);


     /* remove Cilk command-line arguments from argv */
   stop:
     for (j = i; j <= argc; j++)
	  argv[j - (i - 1)] = argv[j];
     argc -= i - 1;

     options->argc = argc;
     options->argv = argv;

     if (options->nmach == 0)
       {
	 /* no -nproc options were specified - run on 1 local process */
	 Cilk_option_machine[0] = NULL;
	 Cilk_option_nproc[0] = DEFAULT_NPROC;
	 options->nmach = 1;
       }

     return 0;
}
