/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk.h>
#include <cilk-internal.h>
#include <cilk-lib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>

/* these variables will be defined in architecture-specific files */
extern int connection;  /* socket for messaging */

#define timer Cilk_get_wall_time
#define tosec Cilk_wall_time_to_sec

#if CILK_DUNIX_DISTRIB
/* no good wall-clock timer - need to use virtual timer */
#undef timer
#define timer Cilk_get_time
#undef tosec
#define tosec Cilk_time_to_sec
#endif

/***********************************************************\
 * Message Passing (UDP implementation)
\***********************************************************/

/* IP address and port number for each machine's UDP socket */
static struct sockaddr_in machinenames[MAXMACHINES];

/* file descriptor for local UDP socket */
static int connection;

/***********************************************************\
 * Initialization
\***********************************************************/

static void print_sockaddr(struct sockaddr_in *name)
{
  CILK_ASSERT(name->sin_family == 2);
  printf("%u.%u.%u.%u(%u)\n",
	 (int)(name->sin_addr.s_addr >>  0) & 0xFF,
	 (int)(name->sin_addr.s_addr >>  8) & 0xFF,
	 (int)(name->sin_addr.s_addr >> 16) & 0xFF,
	 (int)(name->sin_addr.s_addr >> 24) & 0xFF,
	 name->sin_port);
  fflush(stdout);
}

void modify_socket(int sock)
{
  int flags;
  int res;
  
#if 0
  int val;
  /*
   * Set socket buffers to 64K
   */
  val = 64*1024;
  res = setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char*)&val, sizeof(int));
  if (res<0) perror("bad write buf");
  
  val = 64*1024;
  res = setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char*)&val, sizeof(int));
  if (res<0) perror("bad read buf");
#endif
  
  /*
   * Set socket to be non-blocking
   */
  flags = fcntl(sock, F_GETFL);
  CILK_ASSERT(flags >= 0);
  
  flags |= O_NONBLOCK;
  
  res = fcntl(sock, F_SETFL, flags);
  CILK_ASSERT(res >= 0);
}

void Cilk_message_passing_init_machine_globalproc(Options *options, int *mconn)
{
  int i,j;
  int res;
  struct sockaddr_in machines[MAXMACHINES];
  
  /* Read UDP socket name from each machine */
  for(i=0; i<options->nmach; i++)
    {
      res = read(mconn[i], &machines[i], 16);
      if (res != 16)
	return;
      printf("machine %d is ", i); print_sockaddr(&machines[i]);
    }
  
  /* Write all UDP socket names to each machine */
  for(i=0; i<options->nmach; i++)
    {
      for(j=0; j<options->nmach; j++)
	write(mconn[i], &machines[j], 16);
    }
}

static void init_message_buffers(const Options *options);

void Cilk_message_passing_init_machine(const Options *options)
{
  int len = 16;
  struct sockaddr_in name;
  int i;
  int res;
  
  /*
   * Allocate a UDP socket.
   */
  res = getsockname(0, (struct sockaddr *)&name, &len);
  if (res < 0) { perror("can't get socket name!"); exit(1); }
  CILK_ASSERT(len == 16);
  /* Cilk_dprintf_machine("machine default socket  ");print_sockaddr(&name); */
  
  connection = socket(PF_INET, SOCK_DGRAM, 0);
  if (connection < 0) { perror("can't open socket!"); exit(1); }
  
  /* same as default socket, but new port # */
  name.sin_port = 0;
  
  res = bind(connection, (struct sockaddr *)&name, len);
  if (res < 0) { perror("can't bind!"); exit(1); }
  
  res = getsockname(connection, (struct sockaddr *)&name, &len);
  if (res < 0) { perror("can't get socket name!"); exit(1); }
  CILK_ASSERT(len == 16);
  /* Cilk_dprintf_machine("remote listener  ");print_sockaddr(&name); */
  
  /* send name of UDP socket to global process */
  write(1, &name, 16);
  
  /* receive names of other UDP sockets from global process */
  for(i=0; i<options->nmach; i++)
    {
      res = read(0, &machinenames[i], 16);
      if (res != 16) { perror("can't read socket name!"); exit(1); }
    }
  
  /* set local socket parameters */
  modify_socket(connection);
  
  init_message_buffers(options);
}

void Cilk_message_passing_init_local(const Options *options)
{
}


/***********************************************************\
 * UDP messaging
\***********************************************************/

#define SEND_TIMEOUT .030 /* 30 ms */
#define RESEND_REQ_TIMEOUT .005 /* 5 ms */

/* in local time units */
DECLARE_MACHINE_VARIABLE(static Cilk_time, send_timeout, );
DECLARE_MACHINE_VARIABLE(static Cilk_time, resend_req_timeout, );

/*
 * Just big enough to hold one fetch reply (for 8K pages)
 */
#define CHUNK_SIZE (8192+16)

static inline int size_to_chunks(int size)
{
  return (size + CHUNK_SIZE - 1) / CHUNK_SIZE;
}

DECLARE_MACHINE_VARIABLE(static Cilk_mutex, msg_lock, );


/**************     sender side     ***************/
/*
 * Queue of messages to be sent to each other machine
 */
DECLARE_MACHINE_VARIABLE(static SendMessage *, send_msgs, [MAXMACHINES]);

/*
 * Last sucessfully acknowledged sequence number from each receiver
 */
DECLARE_MACHINE_VARIABLE(static int, outbound_ackd_seq_no, [MAXMACHINES]);

/*
 * Last assigned sequence number to each receiver
 */
DECLARE_MACHINE_VARIABLE(static int, outbound_assigned_seq_no, [MAXMACHINES]);

/*
 * Circular buffer of packets waiting to be sent
 */
DECLARE_MACHINE_VARIABLE(static SendReq, send_queue, [MAX_SEND_QUEUE]);
DECLARE_MACHINE_VARIABLE(static unsigned, send_queue_head, );
DECLARE_MACHINE_VARIABLE(static unsigned, send_queue_tail, );

/*
 * Time of last successful message send to each receiver.
 * (messages, not acknowledgements or resend requests)
 */
DECLARE_MACHINE_VARIABLE(static Cilk_time, last_activity, [MAXMACHINES]);


/**************     receiver side     ***************/
/*
 * Queue of messages being received from each other machine
 */
DECLARE_MACHINE_VARIABLE(static RecvMessage *, recv_msgs, [MAXMACHINES]);

/*
 * Highest sequence number received from each sender
 */
DECLARE_MACHINE_VARIABLE(static int, inbound_last_seq_no, [MAXMACHINES]);

/*
 * Last sequence number we can acknowledge to each sender
 * (smaller than above if there are some missing packets)
 */
DECLARE_MACHINE_VARIABLE(static int, inbound_ackable_seq_no, [MAXMACHINES]);

/*
 * List of packets we are expecting but haven't arrived yet
 */
DECLARE_MACHINE_VARIABLE(static MissingPacket, missing_packets, [MAX_MISSING]);
DECLARE_MACHINE_VARIABLE(static int, num_missing_packets, );


/*
 * Statistics stuff
 */
DECLARE_MACHINE_VARIABLE(static int, bytes, );


static void init_message_buffers(const Options *options)
{
  int i;
  for(i=0; i<USE_MACHINE(Machines); i++)
    USE_MACHINE(recv_msgs)[i] = NULL;
  for(i=0; i<USE_MACHINE(Machines); i++)
    USE_MACHINE(send_msgs)[i] = NULL;
  Cilk_mutex_init(&USE_MACHINE(msg_lock));
  for(i=0; i<USE_MACHINE(Machines); i++)
    {
      USE_MACHINE(outbound_ackd_seq_no)[i] = 0;
      USE_MACHINE(outbound_assigned_seq_no)[i] = 0;
      USE_MACHINE(inbound_ackable_seq_no)[i] = 0;
      USE_MACHINE(inbound_last_seq_no)[i] = 0;
    }
  USE_MACHINE(num_missing_packets) = 0;
  USE_MACHINE(send_queue_head) = 0;
  USE_MACHINE(send_queue_tail) = 0;
  
  USE_MACHINE(bytes) = 0;
  
  /* figure out send timeout and resend request timeout, in local time units */
  /* only accurate to a factor of 2, but who cares? */
  i = 1;
  while(tosec(i) < SEND_TIMEOUT)
    i *= 2;
  USE_MACHINE(send_timeout) = i;
  
  i = 1;
  while(tosec(i) < RESEND_REQ_TIMEOUT)
    i *= 2;
  USE_MACHINE(resend_req_timeout) = i;
  
  /* Cilk_dprintf_machine("send: %d  resend: %d\n", USE_MACHINE(send_timeout), USE_MACHINE(resend_req_timeout)); */
}

static inline void add_to_send_queue(SendMessage *s)
{
  SendMessage **entry = &USE_MACHINE(send_msgs)[s->dest];
  while(*entry != NULL)
    {
      entry = &(*entry)->next;
    }
  s->next = NULL;
  *entry = s;
}

static inline int min(int a, int b)
{
  return a<b ? a:b;
}

/*
 * Try to send a packet in the send queue.  Requires that msg_lock
 * be held on entry.  Releases msg_lock on exit.
 */

static void try_to_send(void)
{
  SendMessage *s;
  SendReq *sr;
  TransitHeader header;
  struct msghdr msg;
  int res;
  struct iovec iov[2];
  int seq_no;
  int chunk;
  unsigned x;
  int dest;
  
  CILK_ASSERT(USE_MACHINE(msg_lock) == 1);
  
  if (USE_MACHINE(send_queue_head) == USE_MACHINE(send_queue_tail))
    {
      Cilk_mutex_signal(&USE_MACHINE(msg_lock));
      return;
    }
  
  x = USE_MACHINE(send_queue_head);
  sr = &USE_MACHINE(send_queue)[x];
  seq_no = sr->seq_no;
  s = sr->msg;
  dest = s->dest;
  USE_MACHINE(send_queue_head) = (x+1) % MAX_SEND_QUEUE;
  
  chunk = seq_no - s->seq_no_of_first_chunk;
  CILK_ASSERT(chunk >= 0 && chunk < size_to_chunks(s->size));
  
  /* Cilk_dprintf("trying message %p %d\n", s->handler, chunk); */
  /* build transit header for message send */
  header.handler = s->handler;
  header.size = s->size;
  header.first_seq_no = s->seq_no_of_first_chunk;
  header.sender = USE_MACHINE(Machine);
  header.seq_no = seq_no;
  header.ack_seq_no = USE_MACHINE(inbound_ackable_seq_no)[dest];
  
  /* header for sendmsg */
  msg.msg_name = (char*)&machinenames[dest];
  msg.msg_namelen = 16;
  msg.msg_iov = iov;
  msg.msg_iovlen = 2;
#if CILK_LINUX_DISTRIB
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
#else
  msg.msg_accrights = NULL;
  msg.msg_accrightslen = 0;
#endif
  iov[0].iov_base = (void*)&header;
  iov[0].iov_len = sizeof(TransitHeader);
  iov[1].iov_base = s->data + chunk*CHUNK_SIZE;
  iov[1].iov_len = min(s->size - chunk*CHUNK_SIZE, CHUNK_SIZE);
  
  USE_MACHINE(last_activity)[dest] = timer();
  Cilk_mutex_signal(&USE_MACHINE(msg_lock));
  
  res = sendmsg(connection, &msg, 0);
  /* Cilk_dprintf("sent %d\n", res); */
  if (res == -1)
    {
      if (errno == EWOULDBLOCK)
	{
	  /* splice message chunk back into queue (if not ack'd yet) */
	  Cilk_mutex_wait(&USE_MACHINE(msg_lock));
	  if (seq_no > USE_MACHINE(outbound_ackd_seq_no)[dest])
	    {
	      x = USE_MACHINE(send_queue_head);
	      x = (x+MAX_SEND_QUEUE-1) % MAX_SEND_QUEUE;
	      CILK_ASSERT(x != USE_MACHINE(send_queue_tail));
	      USE_MACHINE(send_queue_head) = x;
	      sr = &USE_MACHINE(send_queue)[x];
	      sr->seq_no = seq_no;
	      sr->msg = s;
	    }
	  Cilk_mutex_signal(&USE_MACHINE(msg_lock));
	  /* Cilk_dprintf("sendmsg in try_to_send() would block!\n"); */
	}
      else
	{
	  perror(NULL);
	  Cilk_die("bad error in try_to_send() %d\n", errno);
	}
    }
  else
    {
      CILK_ASSERT(res == sizeof(TransitHeader) + iov[1].iov_len);
    }
}

void Cilk_rpc_with_size(int dest_pn, void (*handler)(), void *data, int buf_size, int size)
{
  SendMessage *s;
  int i;
  
  /* short cut if destination is the local machine */
  if (dest_pn == USE_MACHINE(Machine))
    {
      handler(data);
      Cilk_free_machine(data, buf_size);
      return;
    }
  
  Cilk_mutex_wait(&USE_MACHINE(msg_lock));
  
  /* initialize message structure and put it in send queue */
  s = Cilk_malloc_machine(sizeof(SendMessage));
  s->handler = handler;
  s->size = size;
  s->buf_size = buf_size;
  s->seq_no_of_first_chunk = USE_MACHINE(outbound_assigned_seq_no)[dest_pn] + 1;
  s->dest = dest_pn;
  s->data = data;
  
  USE_MACHINE(outbound_assigned_seq_no)[dest_pn] += size_to_chunks(size);
  
  add_to_send_queue(s);
  
  /* add each chunk of message to list of chunks to send */
  for(i=0; i<size_to_chunks(size); i++)
    {
      unsigned x;
      SendReq *sr;
      x = USE_MACHINE(send_queue_tail);
      USE_MACHINE(send_queue_tail) = (x+1)%MAX_SEND_QUEUE;
      CILK_ASSERT(USE_MACHINE(send_queue_tail) != USE_MACHINE(send_queue_head));
      sr = &USE_MACHINE(send_queue)[x];
      sr->seq_no = s->seq_no_of_first_chunk + i;
      sr->msg = s;
    }
  
  try_to_send();
}

void Cilk_rpc(int dest_pn, void (*handler)(), void *data, int size)
{
  Cilk_rpc_with_size(dest_pn, handler, data, size, size);
}

static RecvMessage *find_recvmsg(TransitHeader *t)
{
  RecvMessage *r;
  
  for(r = USE_MACHINE(recv_msgs)[t->sender]; r != NULL; r=r->next)
    {
      if (t->seq_no >= r->seq_no_of_first_chunk &&
	  t->seq_no < r->seq_no_of_first_chunk + size_to_chunks(r->size))
	return r;
    }
  
  r = (RecvMessage*)Cilk_malloc_machine(sizeof(RecvMessage));
  r->size = t->size;
  r->seq_no_of_first_chunk = t->first_seq_no;
  r->chunks_so_far = 0;
  r->data = Cilk_malloc_machine(t->size);
  
  r->next = USE_MACHINE(recv_msgs)[t->sender];
  USE_MACHINE(recv_msgs)[t->sender] = r;
  
  return r;
}

static void remove_recvmsg(RecvMessage *r, int sender)
{
  RecvMessage **entry;
  entry = &USE_MACHINE(recv_msgs)[sender];
loop:
  if (*entry == r)
    {
      *entry = r->next;
      return;
    }
  CILK_ASSERT(*entry != NULL);
  entry = &(*entry)->next;
  goto loop;
}

static void free_old_send_messages(int dest)
{
  SendMessage **entry, *s;
  int seq_no = USE_MACHINE(outbound_ackd_seq_no)[dest];
  unsigned x,y;
  
  CILK_ASSERT(USE_MACHINE(msg_lock) == 1);
  
  /* remove old entries from send queue */
  x = USE_MACHINE(send_queue_head);
  y = USE_MACHINE(send_queue_tail);
  while(x != y)
    {
      SendReq *sr = &USE_MACHINE(send_queue)[x];
      if (sr->seq_no <= seq_no && sr->msg->dest == dest)
	{
	  y = (y+MAX_SEND_QUEUE-1) % MAX_SEND_QUEUE;
	  *sr = USE_MACHINE(send_queue)[y];
	}
      else
	{
	  x = (x+1) % MAX_SEND_QUEUE;
	}
    }
  USE_MACHINE(send_queue_tail) = y;
  
  /* remove and deallocate old messages */
  entry = &USE_MACHINE(send_msgs)[dest];
loop:
  s = *entry;
  if (s == NULL) return;
  if (seq_no >= s->seq_no_of_first_chunk + size_to_chunks(s->size) - 1)
    {
      *entry = s->next;
      Cilk_free_machine(s->data, s->buf_size);
      Cilk_free_machine(s, sizeof(SendMessage));
    }
  else
    {
      entry = &s->next;
    }
  goto loop;
}

static void run_handler(void (*handler)(), void *data)
{
#if CILK_TIMING
  void Cilk_uninclude_time(Cilk_time process, Cilk_time wall);
  Cilk_time begin = Cilk_get_time();
  Cilk_time begin_wall = Cilk_get_wall_time();
#endif
  
  handler(data);
  
#if CILK_TIMING
  Cilk_uninclude_time(Cilk_get_time() - begin,
		      Cilk_get_wall_time() - begin_wall);
#endif
}

static void send_ack(int dest)
{
  TransitHeader header;
  struct msghdr msg;
  struct iovec iov[1];
  int res;
  
  /* Cilk_dprintf("sending ack for %d\n", USE_MACHINE(inbound_ackable_seq_no)[dest]); */
  
  header.handler = NULL;
  header.size = 0;
  header.first_seq_no = 0;
  header.sender = USE_MACHINE(Machine);
  header.seq_no = 0;   /* ensures ack message is discarded */
  header.ack_seq_no = USE_MACHINE(inbound_ackable_seq_no)[dest];
  /* note inbound_ackable_seq_no doesn't need to be protected */
  
  /* header for sendmsg */
  msg.msg_name = (char*)&machinenames[dest];
  msg.msg_namelen = 16;
  msg.msg_iov = iov;
  msg.msg_iovlen = 1;
#if CILK_LINUX_DISTRIB
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
#else
  msg.msg_accrights = NULL;
  msg.msg_accrightslen = 0;
#endif
  iov[0].iov_base = (void*)&header;
  iov[0].iov_len = sizeof(TransitHeader);
  
  res = sendmsg(connection, &msg, 0);
  
  if (res == -1)
    {
      if (errno != EWOULDBLOCK)
	{
	  perror(NULL);
	  Cilk_die("bad error in send_ack()\n");
	}
    }
  else
    {
      CILK_ASSERT(res == sizeof(TransitHeader));
    }
}

static void send_resend_req(int dest, int seq_no)
{
  TransitHeader header;
  struct msghdr msg;
  struct iovec iov[1];
  int res;
  
  /* Cilk_dprintf("sending resend req for %d to %d\n", seq_no, dest); */
  
  header.handler = NULL;
  header.size = 2;              /* size of 2 encodes resend request */
  header.first_seq_no = seq_no;
  header.sender = USE_MACHINE(Machine);
  header.seq_no = 0;   /* ensures ack message is discarded */
  header.ack_seq_no = USE_MACHINE(inbound_ackable_seq_no)[dest];
  
  /* header for sendmsg */
  msg.msg_name = (char*)&machinenames[dest];
  msg.msg_namelen = 16;
  msg.msg_iov = iov;
  msg.msg_iovlen = 1;
#if CILK_LINUX_DISTRIB
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
#else
  msg.msg_accrights = NULL;
  msg.msg_accrightslen = 0;
#endif
  iov[0].iov_base = (void*)&header;
  iov[0].iov_len = sizeof(TransitHeader);
  
  Cilk_mutex_signal(&USE_MACHINE(msg_lock));
  
  res = sendmsg(connection, &msg, 0);
  
  if (res == -1)
    {
      if (errno != EWOULDBLOCK)
	{
	  perror(NULL);
	  Cilk_die("bad error in send_resend_req()\n");
	}
    }
  else
    {
      CILK_ASSERT(res == sizeof(TransitHeader));
    }
}

static void resend(int seq_no, int dest)
{
  unsigned x;
  SendReq *sr;
  SendMessage *s;
  
  CILK_ASSERT(USE_MACHINE(msg_lock) == 1);
  CILK_ASSERT(seq_no > USE_MACHINE(outbound_ackd_seq_no)[dest]);
  
  /* find which message has this sequence number */
  for(s = USE_MACHINE(send_msgs)[dest]; s != NULL; s=s->next)
    {
      if (seq_no >= s->seq_no_of_first_chunk &&
	  seq_no < s->seq_no_of_first_chunk + size_to_chunks(s->size))
	break;
    }
  if (s == NULL)
    {
      Cilk_dprintf("attempted resend of already ackd packet %d (ackd=%d, assigned=%d)\n", seq_no, USE_MACHINE(outbound_ackd_seq_no)[dest], USE_MACHINE(outbound_assigned_seq_no)[dest]);
    }
  CILK_ASSERT(s != NULL);
  
  /* put resend request at head of send queue */
  x = USE_MACHINE(send_queue_head);
  x = (x+MAX_SEND_QUEUE-1) % MAX_SEND_QUEUE;
  CILK_ASSERT(x != USE_MACHINE(send_queue_tail));
  USE_MACHINE(send_queue_head) = x;
  sr = &USE_MACHINE(send_queue)[x];
  sr->seq_no = seq_no;
  sr->msg = s;
}

static void got_chunk(TransitHeader *t)
{
  RecvMessage *r;
  void *chunk_data = (void*)(t + 1);
  int chunk;
  int amount;
  void *msg_data;
  
  Cilk_mutex_wait(&USE_MACHINE(msg_lock));
  
  /*
   * Take acknowledgement out of packet, free messages older
   * than acknowledgement
   */
  if (t->ack_seq_no > USE_MACHINE(outbound_ackd_seq_no)[t->sender])
    {
      USE_MACHINE(outbound_ackd_seq_no)[t->sender] = t->ack_seq_no;
      free_old_send_messages(t->sender);
    }
  
  /*
   * Old packet
   */
  if (t->seq_no <= USE_MACHINE(inbound_last_seq_no)[t->sender])
    {
      if (t->seq_no == 0)
	{
	  /*
	   * Sequence #0 packets are special
	   */
	  switch(t->size)
	    {
	    case 0:
	      /* Cilk_dprintf("got ack packet %d\n", t->ack_seq_no) */;
	      goto done;
	      
	    case 2:  /* resend request */
	      /* first_seq_no holds sequence number to resend */
	      
	      /* Cilk_dprintf("got resend request for %d from %d\n", t->first_seq_no, t->sender); */
	      if (t->first_seq_no > USE_MACHINE(outbound_ackd_seq_no)[t->sender])
		resend(t->first_seq_no, t->sender);
#if 0
	      else
		Cilk_dprintf("old resend request %d\n", t->first_seq_no);
#endif
	      break;
	    }
	}
      else
	{
	  /*
	   * Check missing packet list
	   */
	  int i;
	  for(i=0; i<USE_MACHINE(num_missing_packets); i++)
	    {
	      MissingPacket *p = &USE_MACHINE(missing_packets)[i];
	      if (p->sender == t->sender && p->seq_no == t->seq_no)
		{
		  int seq_no;
		  
		  /* remove from missing packet list */
		  *p = USE_MACHINE(missing_packets)[--USE_MACHINE(num_missing_packets)];
		  
		  /* update inbound_ackable_seq_no */
		  seq_no = USE_MACHINE(inbound_last_seq_no)[t->sender];
		  for(i=0; i<USE_MACHINE(num_missing_packets); i++)
		    {
		      MissingPacket *p = &USE_MACHINE(missing_packets)[i];
		      if (p->sender == t->sender && p->seq_no <= seq_no)
			seq_no = p->seq_no-1;
		    }
		  USE_MACHINE(inbound_ackable_seq_no)[t->sender] = seq_no;
		  
		  /* Cilk_dprintf("got missing packet %d from %d\n", t->seq_no, t->sender); */
		  
		  /* process packet as normal */
		  goto process_chunk;
		}
	    }
	  /* Cilk_dprintf("got old packet %d from %d\n", t->seq_no, t->sender); */
	  /* send_ack(t->sender); */
	}
      goto done;
    }
  
  /*
   * Sequence number is too new - record missing packets
   */
  if (t->seq_no > 1 + USE_MACHINE(inbound_last_seq_no)[t->sender])
    {
      int i;
      /*
       * Add intervening packets to missing packet list
       */
      for(i=USE_MACHINE(inbound_last_seq_no)[t->sender]+1; i<t->seq_no; i++)
	{
	  MissingPacket *p = &USE_MACHINE(missing_packets)[USE_MACHINE(num_missing_packets)++];
	  CILK_ASSERT(USE_MACHINE(num_missing_packets) < MAX_MISSING);
	  
	  /* Cilk_dprintf("adding missing packet %d from %d\n", i, t->sender); */
	  
	  p->sender = t->sender;
	  p->seq_no = i;
	  p->gone_missing = timer();
	}
    }
  
  /*
   * Update last sequence number
   */
  USE_MACHINE(inbound_last_seq_no)[t->sender] = t->seq_no;
  if (USE_MACHINE(inbound_ackable_seq_no)[t->sender] == t->seq_no - 1)
    USE_MACHINE(inbound_ackable_seq_no)[t->sender] = t->seq_no;
  
process_chunk:
  /* do a special case for a one-chunk message */
  if (t->size <= CHUNK_SIZE)
    {
      USE_MACHINE(bytes) += t->size;
      Cilk_mutex_signal(&USE_MACHINE(msg_lock));
      
      send_ack(t->sender);
      
      run_handler(t->handler, chunk_data);
      return;
    }
  
  r = find_recvmsg(t);
  
  CILK_ASSERT(r->size == t->size);
  
  chunk = t->seq_no - r->seq_no_of_first_chunk;
  msg_data = r->data + chunk * CHUNK_SIZE;
  amount = min(t->size - chunk * CHUNK_SIZE, CHUNK_SIZE);
  USE_MACHINE(bytes) += amount;
  
  Cilk_mutex_signal(&USE_MACHINE(msg_lock));
  
  /* Cilk_dprintf("received chunk %d of %d\n", (t->seq_no - r->seq_no_of_first_chunk), size_to_chunks(t->size)); */
  
  /*
   * Send acknowledgement
   */
  send_ack(t->sender);
  
  /* copy chunk data into message */
  memcpy(msg_data, chunk_data, amount);
  
  Cilk_mutex_wait(&USE_MACHINE(msg_lock));
  
  r->chunks_so_far++;
  if(r->chunks_so_far == size_to_chunks(t->size))
    {
      remove_recvmsg(r, t->sender);
      Cilk_mutex_signal(&USE_MACHINE(msg_lock));
      
      /* run handler function */
      run_handler(t->handler, r->data);
      
      Cilk_free_machine(r->data, r->size);
      Cilk_free_machine(r, sizeof(RecvMessage));
      
      return;
    }
done:
  Cilk_mutex_signal(&USE_MACHINE(msg_lock));
}

void Cilk_poll(void)
{
  struct msghdr msg;
  int res;
  struct iovec iov[1];
  char buf[sizeof(TransitHeader)+CHUNK_SIZE];
  int i;
  
  /*
   * Try to receive messages on input socket
   */
  while(1)
    {
      /* header for receiving */
      msg.msg_name = NULL;
      msg.msg_namelen = 0;
      msg.msg_iov = iov;
      msg.msg_iovlen = 1;
#if CILK_LINUX_DISTRIB
      msg.msg_control = NULL;
      msg.msg_controllen = 0;
#else
      msg.msg_accrights = NULL;
      msg.msg_accrightslen = 0;
#endif
      iov[0].iov_base = buf;
      iov[0].iov_len = sizeof(buf);
      
      res = recvmsg(connection, &msg, 0);
      /* Cilk_dprintf("received %d\n", res); */
      
      if (res == -1)
	{
	  if (errno == EWOULDBLOCK)
	    {
	      /* Cilk_dprintf("recv would block\n"); */
	    }
	  else
	    {
	      perror(NULL);
	      Cilk_die("bad result in Cilk_poll\n");
	    }
	  break;
	}
      else
	{
	  got_chunk((TransitHeader*)buf);
	}
    }
  
  /*
   * Request retransmit on missing packets
   */
  if (USE_MACHINE(num_missing_packets) > 0)
    if (Cilk_mutex_try(&USE_MACHINE(msg_lock)))
      {
	for(i=0; i<USE_MACHINE(num_missing_packets); i++)
	  {
	    MissingPacket *p = &USE_MACHINE(missing_packets)[i];
	    if (p->gone_missing + USE_MACHINE(resend_req_timeout) < timer())
	      {
		p->gone_missing = timer();
		/* request resend of p->seq_no from p->sender */
		send_resend_req(p->sender, p->seq_no);  /* releases msg_lock */
		goto push;
	      }
	  }
	Cilk_mutex_signal(&USE_MACHINE(msg_lock));
      }
  
  /*
   * Try to push any unsent or partially sent messages
   */
push:
  if (Cilk_mutex_try(&USE_MACHINE(msg_lock)))
    {
      try_to_send();
      if (!Cilk_mutex_try(&USE_MACHINE(msg_lock)))
	return;
      /*
       * Check for no activity coupled with unacknowleged packet
       */
      for(i=0; i<USE_MACHINE(Machines); i++)
	{
	  int seq_no;
	  
	  if (i == USE_MACHINE(Machine))
	    continue;
	  if (USE_MACHINE(send_msgs)[i] == NULL)
	    continue;
	  CILK_ASSERT(USE_MACHINE(outbound_ackd_seq_no)[i] < USE_MACHINE(outbound_assigned_seq_no)[i]);
	  if (USE_MACHINE(last_activity)[i] + USE_MACHINE(send_timeout) > timer())
	    continue;
	  
	  USE_MACHINE(last_activity)[i] = timer();
	  
	  /*
	   * Send last assigned packet we can again.
	   */
	  seq_no = USE_MACHINE(outbound_assigned_seq_no)[i];
	  
	  /* Cilk_dprintf("doing timeout resend of %d to %d\n", seq_no, i); */
	  /* Cilk_dprintf("out_ackd %d out_assn %d\n", USE_MACHINE(outbound_ackd_seq_no)[i], USE_MACHINE(outbound_assigned_seq_no)[i]); */
	  
	  resend(seq_no, i);
	}
      
      Cilk_mutex_signal(&USE_MACHINE(msg_lock));
    }
}

void *Cilk_alloc_msg_buf(int size, int dest_pn)
{
  return Cilk_malloc_machine(size);
}

void Cilk_free_msg_buf(void *data, int size, int dest_pn)
{
  Cilk_free_machine(data, size);
}

void Cilk_message_passing_init_global(Options *options)
{
}
