
/*
 * $Id: stats.c,v 1.2.2.3 1997/11/10 22:28:01 randall Exp $
 */

/*
 * $Log: stats.c,v $
 * Revision 1.2.2.3  1997/11/10 22:28:01  randall
 * Updated some interfaces, moved some more socket initialization
 * into message-udp.c
 *
 * Revision 1.2.2.2  1997/10/14 20:26:13  randall
 * First working version of distributed cilk on DEC machines.
 *
 * Revision 1.2.2.1  1997/06/13 22:33:33  randall
 * Added distributed version to the repository under
 * branch "distributed"
 *
 * Revision 1.2  1997/03/01 17:42:57  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/***********************************************************\
 * Statistics gathering
\***********************************************************/

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>

#if CILK_STATS
#include <stdio.h>

typedef struct Stats {
     long event_count[EVENT_NTYPES];
} Stats;

/* valid only on processor 0 */
DECLARE_MACHINE_VARIABLE(static struct Stats *, stat_array, );

/*
 * Because of perl script hack, we can't use Stats* here, even
 * though that is what we want.
 */
DECLARE_PRIVATE_VARIABLE(static long *, stats, );

/* valid only on machine 0 */
DECLARE_MACHINE_VARIABLE(static Cilk_mutex, recv_stat_lock, );
DECLARE_MACHINE_VARIABLE(static volatile int, recv_stat_counter, );

/*
 * table of description of events. It is printed in
 * the same order.
 */
static const struct event_desc {
     char *name;
     int event;
     int verbosity;   /* if statslevel >= 2 + verbosity this thing is printed */
} event_descriptions[] = {
     {
	  "Threads", EVENT_THREAD,               0
     },
#if 0
     /* this is the same as # of steals, if there are no bugs */
     {
	  "Closures", EVENT_CLOSURE_CREATE,      0
     },
#endif
     {
	  "Steals", EVENT_STEAL,                 0
     },
     {
	  "Attempts", EVENT_STEAL_ATTEMPT,       1
     },
     {
	  "Fail/EmptyQ", EVENT_STEAL_EMPTY_DEQUE,  4
     },
     {
	  "Fail/Dekker", EVENT_STEAL_NO_DEKKER,    4
     },
     {
	  "Fail/Return", EVENT_STEAL_RETURNING,    4
     },
     {
	  "Fail/Abort", EVENT_STEAL_ABORT,         4
     },
     {
	  "Abort/Slow", EVENT_ABORT_SLOW,          0
     },
     {
	  "Abort/Alone", EVENT_ABORT_STANDALONE,   1
     },
     {
	  "Ready->Susp", EVENT_ABORT_READY_SUSPEND,  4
     },
     {
	  "Ready->Ret", EVENT_ABORT_READY_RETURN,    4
     },
     {
	  "Prov. good", EVENT_PROVABLY_GOOD_STEAL,   4
     },
     {
	  "Suspend", EVENT_SUSPEND,                  1
     },
     {
	  "Sync check", EVENT_CILK_SYNC,             1
     },
     {
	  "Exceptions", EVENT_EXCEPTION,             1
     },
     {
	  "Ex/Other", EVENT_EXCEPTION_OTHER,         4
     },
     {
	  "Ex/Steal", EVENT_EXCEPTION_STEAL,         4
     },
     {
	  "Ex/Abort", EVENT_EXCEPTION_ABORT,         4
     },
     {
	  "Ex/Abort_ret", EVENT_EXCEPTION_ABORT_RETURN, 4
     },
     {
	  "Ex/Abort_susp", EVENT_EXCEPTION_ABORT_SUSPEND, 4
     },
     {
	  "Return/slow", EVENT_RETURN_SLOW,          4
     },
     {
	  "Enqueue", EVENT_RETURN_ENQUEUE,           4
     },
     {
	  "Poll", EVENT_POLL_INLETS,                 2
     },
     {
          "Remote steal", EVENT_STEAL_REMOTE_SUCCESS, 0
     },
     {
          "Remote fail", EVENT_STEAL_REMOTE_FAILURE,  0
     },
     {
          "Fetch", EVENT_FETCH,                  0
     },
     {
          "Fetch Zero", EVENT_FETCH_ZERO,        0
     },
     {
          "Fetch Null", EVENT_FETCH_NULL,        0
     },
     {
          "Reconcile", EVENT_RECONCILE,          0
     },
     {
          "Rec Null", EVENT_RECONCILE_NULL,      0
     },
     {
	  "User 0", EVENT_USER0, 0
     },
     {
	  "User 1", EVENT_USER1, 0
     },
     {
	  "User 2", EVENT_USER2, 0
     },
     {
	  "User 3", EVENT_USER3, 0
     },
     {
	  "User 4", EVENT_USER4, 0
     },
     {
	  "User 5", EVENT_USER5, 0
     },
     {
	  "User 6", EVENT_USER7, 0
     },
     {
	  "User 7", EVENT_USER7, 0
     },
     {
	  (char *) 0, 0
     }
};

void Cilk_stats_init_local(const Options *options)
{
  int i;
  USE_PRIVATE(stats) = malloc(sizeof(Stats));
  CILK_CHECK(USE_PRIVATE(stats) != NULL,
	     ("couldn't allocate local stats!\n"));
  for (i = 0; i < EVENT_NTYPES; i++)
    USE_PRIVATE(stats)[i] = 0;
}

void Cilk_stats_init_machine(const Options *options)
{
  int size;
  if (USE_MACHINE(Machine) == 0)
    {
      size = options->nproc * sizeof(Stats);
      USE_MACHINE(stat_array) = Cilk_malloc_machine(size);
      CILK_CHECK(USE_MACHINE(stat_array) != NULL,
		 ("couldn't allocate stat array!\n"));
      
      Cilk_mutex_init(&USE_MACHINE(recv_stat_lock));
      USE_MACHINE(recv_stat_counter) = options->nproc;
    }
}

void Cilk_stats_init_global(Options *options)
{
}

#define NAME_DESC "%13s"
#define FIELD_DESC "%9ld"
#define TOTAL_DESC "%10s"
#define TOTAL_FIELD_DESC "%10ld"
#define PROC_PER_LINE 5

typedef struct
{
  int self;
  Stats stats;
} StatXfer;

void stat_recv(StatXfer *req)
{
  CILK_ASSERT(USE_MACHINE(Machine) == 0);
  
  USE_MACHINE(stat_array)[req->self] = req->stats;
  
  Cilk_fence();
  
  Cilk_mutex_wait(&USE_MACHINE(recv_stat_lock));
  USE_MACHINE(recv_stat_counter)--;
  Cilk_mutex_signal(&USE_MACHINE(recv_stat_lock));
}

void Cilk_stats_done_local(const Options *options)
{
  if (options->statlevel < 2)
    return;
  
  /* send stat data to machine 0 */
  {
    StatXfer *req = Cilk_alloc_msg_buf(sizeof(StatXfer), 0);
    req->self = Self;
    req->stats = *(Stats*)USE_PRIVATE(stats);
    
    Cilk_rpc(0, stat_recv, req, sizeof(*req));
  }
  
  if (Self == 0)
    {
      int i, j;
      int start, end, line;
      Stats total;
      struct event_desc *p;
      FILE *infofile;
      int active_size = options->nproc;
      
      while(USE_MACHINE(recv_stat_counter) != 0)
	Cilk_poll();
      Cilk_fence();
      
#if 0
      if (strcmp(options->infofile_name, "-") == 0)
	infofile = stderr;
      else
	infofile = fopen(options->infofile_name, "w");
#endif
      infofile = stdout;
      
      CILK_CHECK(infofile != NULL,
		 ("can't open stdout for writing\n"));
      
      for (j = 0; j < EVENT_NTYPES; j++)
	total.event_count[j] = 0;
      
      /* compute totals */
      for (i = 0; i < active_size; ++i)
	for (j = 0; j < EVENT_NTYPES; j++)
	  total.event_count[j] += 
	    USE_MACHINE(stat_array)[i].event_count[j];
      
      fprintf(infofile,
	      "\nRUNTIME SYSTEM STATISTICS:\n"
	      "\n");
      
      for (line = 0; 
	   line <= (active_size - 1) / PROC_PER_LINE;
	   ++line)
	{
	  start = line * PROC_PER_LINE;
	  end = (line + 1) * PROC_PER_LINE;
	  if (end > active_size)
	    end = active_size;
	  
	  fprintf(infofile, NAME_DESC, "PN");
	  for (i = start; i < end; ++i)
	    fprintf(infofile, FIELD_DESC, (long) i);
	  
	  if (i == active_size)
	    fprintf(infofile, TOTAL_DESC, "Total");
	  
	  fprintf(infofile, "\n");
	  
	  for (p = (struct event_desc *) event_descriptions; p->name; p++)
	    {
	      fprintf(infofile, NAME_DESC, p->name);
	      for (i = start; i < end; ++i)
		fprintf(infofile, FIELD_DESC,
			USE_MACHINE(stat_array)[i].event_count[p->event]);
	      if (i == active_size)
		fprintf(infofile, TOTAL_FIELD_DESC,
			total.event_count[p->event]);
	      fprintf(infofile, "\n");
	    }
	  
	  fprintf(infofile, "\n");
	  
	}
    }
}

void Cilk_stats_done_machine(const Options *options)
{
  if (USE_MACHINE(Machine) == 0)
    {
      Cilk_free_machine(USE_MACHINE(stat_array), options->nproc * sizeof(Stats));
    }
}
void Cilk_stats_done_global(const Options *options)
{
}

void Cilk_event(int type)
{
  USE_PRIVATE(stats)[type]++;
}

void Cilk_event_new_thread(void)
{
  USE_PRIVATE(stats)[EVENT_THREAD]++;
}

#endif				/* CILK_STATS */
