/*
 * $Id: timing.c,v 1.2.2.6 1997/12/18 23:14:17 randall Exp $
 */

/*
 * $Log: timing.c,v $
 * Revision 1.2.2.6  1997/12/18 23:14:17  randall
 * Cache eviction fix, some Makefile changes.
 *
 * Revision 1.2.2.5  1997/12/05 23:39:56  randall
 * Fixed bug in cache and simplified stuff a bit.
 *
 * Revision 1.2.2.4  1997/12/01 19:34:35  randall
 * Small updates.
 *
 * Revision 1.2.2.3  1997/11/10 22:28:02  randall
 * Updated some interfaces, moved some more socket initialization
 * into message-udp.c
 *
 * Revision 1.2.2.2  1997/10/14 20:26:14  randall
 * First working version of distributed cilk on DEC machines.
 *
 * Revision 1.2.2.1  1997/06/13 22:33:35  randall
 * Added distributed version to the repository under
 * branch "distributed"
 *
 * Revision 1.2  1997/03/01 17:42:59  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/***********************************************************
 * Time-gathering stuff
 ***********************************************************/

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>

#if CILK_TIMING
#include <stdio.h>

/*
 * The model for time gathering is the following:
 * there is a certain set of states in which the system can
 * be. Throughout the code, ENTER_STATE(something) signals
 * that a given state is entered. EXIT_STATE(state) reverts back to
 * the previous state, and checks that the right state is being
 * exited. The time is accumulated on a per-state
 * basis. Time is cumulative, i.e. it is measured from ENTER_STATE
 * to EXIT_STATE, regardless of other states entered in between.
 * 
 * No more than STATE_DEPTH states can be nested.
 * The stack is really used only for checking purposes, but it's
 * probably a good idea to keep it.
 *
 */
#define STATE_DEPTH 256

/* also update name table below */

typedef struct StateInfo {
     volatile int state_stackp;
     volatile int state_stack[STATE_DEPTH];
     volatile Cilk_time state_stack_begintime[STATE_DEPTH];
     volatile Cilk_time statetime[TIME_NSTATES];
     volatile Cilk_time state_stack_begin_wall_time[STATE_DEPTH];
     volatile Cilk_time state_wall_time[TIME_NSTATES];
     volatile int stateentered[TIME_NSTATES];
     CILK_CACHE_LINE_PAD;
} StateInfo;


DECLARE_MACHINE_VARIABLE(static struct StateInfo *, timer_state_array, );
DECLARE_PRIVATE_VARIABLE(static struct StateInfo *, timer_state_info, );

/*
 * Valid only on machine 0
 */
DECLARE_MACHINE_VARIABLE(static struct StateInfo *, timer_array, );
DECLARE_MACHINE_VARIABLE(static Cilk_mutex, timer_lock, );
DECLARE_MACHINE_VARIABLE(static int, timer_count, );

/*
 * table of description of states. It is printed in
 * the same order.
 */
static const struct state_desc {
     char *name;
     int state;
} state_descriptions[] = {

     {
	  "Total execution time", STATE_TOTAL
     },
     {
	  "Working time", STATE_WORKING
     },
     {
	  "Waiting for locks", STATE_WAITING_FOR_LOCK
     },
     {
	  "Stealing", STATE_STEALING
     },
     {
	  "Returning", STATE_RETURNING
     },
     {
        "Aborting", STATE_ABORT_RECURSIVE
     },
     {
          "Diffing", STATE_COMPUTE_DIFF
     },
     {
          "Applying", STATE_APPLY_DIFF
     },
     {
          "Copying", STATE_COPY_PAGE
     },
     {
          "Twinning", STATE_TWIN_PAGE
     },
     {
          "Zeroing", STATE_ZERO_PAGE
     },
     {
	  "Fetching", STATE_WAITING_FOR_FETCH
     },
     {
	  "Reconciling", STATE_WAITING_FOR_RECONCILE
     },
     {
	  "Bstore", STATE_BSTORE_OP
     },
     {
	  "User state 0", STATE_USER0
     },
     {
	  "User state 1", STATE_USER1
     },
     {
	  "User state 2", STATE_USER2
     },
     {
	  "User state 3", STATE_USER3
     },
     {
	  "User state 4", STATE_USER4
     },
     {
	  "User state 5", STATE_USER5
     },
     {
	  "User state 6", STATE_USER6
     },
     {
	  "User state 7", STATE_USER7
     },
     {
	  (char *) 0, 0
     }
};

static char *smart_sprint_time(double x)
{
     static char buf[128];

     if (x < 1.0E-3)
	  sprintf(buf, "%f us", x * 1.0E6);
     else if (x < 1.0)
	  sprintf(buf, "%f ms", x * 1.0E3);
     else
	  sprintf(buf, "%f s", x);

     return buf;
}

void Cilk_enter_state(int state)
{
     int stackp;
     Cilk_time now = Cilk_get_time();
     Cilk_time now_wall = Cilk_get_wall_time();
     StateInfo *info;
     info = USE_PRIVATE(timer_state_info);
     if (info == NULL)
	  return;
     stackp = info->state_stackp;
     CILK_ASSERT(stackp < STATE_DEPTH);
     info->state_stack[stackp] = state;
     info->state_stack_begintime[stackp] = now;
     info->state_stack_begin_wall_time[stackp] = now_wall;
     info->state_stackp = stackp + 1;
     ++(info->stateentered[state]);
}

void Cilk_exit_state(int state)
{
     int stackp;
     Cilk_time now = Cilk_get_time();
     Cilk_time now_wall = Cilk_get_wall_time();
     StateInfo *info;
     info = USE_PRIVATE(timer_state_info);
     if (info == NULL)
	  return;
     stackp = info->state_stackp - 1;
     info->state_stackp = stackp;
     CILK_ASSERT(stackp >= 0);

     CILK_ASSERT(info->state_stack[stackp] == state);
     info->statetime[state] += now - info->state_stack_begintime[stackp];
     info->state_wall_time[state] += 
	  now_wall - info->state_stack_begin_wall_time[stackp];
}

/*
 * Remove given amount of time from all including intervals
 */
void Cilk_uninclude_time(Cilk_time time, Cilk_time wall_time)
{
  int stackp;
  int state;
  int i;
  StateInfo *info;
  info = USE_PRIVATE(timer_state_info);
  
  stackp = info->state_stackp;
  for(i=0; i<stackp; i++)
    {
      state = info->state_stack[i];
      info->statetime[state] -= time;
      info->state_wall_time[state] -= wall_time;
    }
}

/* print time statistics */
static void Cilk_print_time_statistics(int active_size)
{
     int i;
     StateInfo *info = USE_MACHINE(timer_array);
     double worktime = 0.0;
     double totaltime = 0.0;
     FILE *infofile = stdout;

     CILK_ASSERT(USE_MACHINE(Machine) == 0);

     fprintf(infofile,
	     "\nSUMMARY OF TIME SPENT BY EACH PROCESSOR:\n\n"
	     " PN    Working      Total\n"
	     "-------------------------\n");
     for (i = 0; i < active_size; i++) {
	  fprintf(infofile, "%3d%11f%11f\n", i,
	      Cilk_time_to_sec(info[i].statetime[STATE_WORKING]),
	       Cilk_time_to_sec(info[i].statetime[STATE_TOTAL]));
	  worktime +=
	      Cilk_time_to_sec(info[i].statetime[STATE_WORKING]);
	  totaltime +=
	      Cilk_time_to_sec(info[i].statetime[STATE_TOTAL]);
     }
     fprintf(infofile,
	     "-------------------------\n"
	     "TOT%11f%11f\n", worktime, totaltime);
}

static void Cilk_summarize_time_statistics(int active_size)
{
     int i;
     struct state_desc *p;
     StateInfo *info = USE_MACHINE(timer_array);
     FILE *infofile = stdout;

     CILK_ASSERT(USE_MACHINE(Machine) == 0);

     fprintf(infofile,
	     "\nSUMMARY OF TIME STATISTICS:\n\n");

     for (p = (struct state_desc *) state_descriptions; p->name; ++p) {
	  Cilk_time virtual = 0;
	  Cilk_time wall = 0;
	  int entered = 0;

	  for (i = 0; i < active_size; i++) {
	       virtual += info[i].statetime[p->state];
	       wall += info[i].state_wall_time[p->state];
	       entered += info[i].stateentered[p->state];
	  }

	  if (entered > 0)
	       fprintf(infofile,
		       "%22s: %10.6fs / wall %10.6fs (entered %8d times)\n",
		       p->name,
		       Cilk_time_to_sec(virtual),
		       Cilk_wall_time_to_sec(wall),
		       entered);
     }
}

static void Cilk_print_detailed_time_statistics(int active_size)
{
     int i;
     struct state_desc *p;
     StateInfo *info = USE_MACHINE(timer_array);
     FILE *infofile = stdout;

     CILK_ASSERT(USE_MACHINE(Machine) == 0);

     fprintf(infofile,
	     "\nDETAILS OF TIME SPENT BY EACH PROCESSOR:\n");

     for (i = 0; i < active_size; i++) {
	  fprintf(infofile,
		  "\n--- Time statistics for processor %d ---\n", i);

	  for (p = (struct state_desc *) state_descriptions; p->name; ++p) {
	       if (info[i].stateentered[p->state] > 0)
		    fprintf(infofile,
			    "%22s: %10.6fs / wall %10.6fs (entered %8d times)\n",
			    p->name,
			    Cilk_time_to_sec(info[i].statetime
						     [p->state]),
			    Cilk_wall_time_to_sec(info[i].state_wall_time
						  [p->state]),
			    info[i].stateentered[p->state]);
	  }
     }
}

double Cilk_compute_work(int active_size)
{
     int i;
     double work = 0.0;
     StateInfo *info = USE_MACHINE(timer_array);

     CILK_ASSERT(USE_MACHINE(Machine) == 0);

     for (i = 0; i < active_size; i++)
	  work += Cilk_time_to_sec(info[i].statetime[STATE_WORKING]);

     return work;
}

void Cilk_timing_init_local(const Options *options)
{
  int i;
  USE_PRIVATE(timer_state_info) = &USE_MACHINE(timer_state_array)[USE_PRIVATE(Process)];
  
  for (i = 0; i < TIME_NSTATES; i++)
    {
      USE_PRIVATE(timer_state_info)->statetime[i] = (Cilk_time) 0;
      USE_PRIVATE(timer_state_info)->stateentered[i] = 0;
    }
  USE_PRIVATE(timer_state_info)->state_stackp = 0;
}

void timing_reset(void)
{
  int p, i;
  Cilk_time now = Cilk_get_time();
  Cilk_time now_wall = Cilk_get_wall_time();
  for(p = 0; p < USE_MACHINE(Processes); p++)
    {
      struct StateInfo *info = &USE_MACHINE(timer_state_array)[p];
      for (i = 0; i < TIME_NSTATES; i++)
	{
	  info->statetime[i] = (Cilk_time) 0;
	  info->state_wall_time[i] = (Cilk_time) 0;
	  info->stateentered[i] = 0;
	}
      for(i = 0; i < info->state_stackp; i++)
	{
	  info->state_stack_begintime[i] = now;
	  info->state_stack_begin_wall_time[i] = now_wall;
	}
    }
}

void Cilk_timing_init_machine(const Options *options)
{
  if (USE_MACHINE(Machine) == 0)
    {
      USE_MACHINE(timer_array) = Cilk_malloc_machine(options->nproc * sizeof(StateInfo));
      Cilk_mutex_init(&USE_MACHINE(timer_lock));
      USE_MACHINE(timer_count) = options->nproc;
    }
  
  USE_MACHINE(timer_state_array) = Cilk_malloc_machine(options->permach * sizeof(StateInfo));
  CILK_CHECK(USE_MACHINE(timer_state_array) != NULL,
	     ("couldn't allocate local timer info!\n"));
}

void Cilk_timing_init_global(Options *options)
{
}

typedef struct
{
  int self;
  StateInfo timing;
} TimingXfer;

void timing_recv(TimingXfer *req)
{
  CILK_ASSERT(USE_MACHINE(Machine) == 0);
  
  USE_MACHINE(timer_array)[req->self] = req->timing;
  
  Cilk_fence();
  
  Cilk_mutex_wait(&USE_MACHINE(timer_lock));
  USE_MACHINE(timer_count)--;
  Cilk_mutex_signal(&USE_MACHINE(timer_lock));
}

void Cilk_timing_done_local(const Options *options)
{
  {
    TimingXfer *req = Cilk_alloc_msg_buf(sizeof(TimingXfer), 0);
    req->self = Self;
    req->timing = *(StateInfo*)USE_PRIVATE(timer_state_info);
    
    Cilk_rpc(0, timing_recv, req, sizeof(*req));
  }
  
  if (Self == 0)
    {
      while(USE_MACHINE(timer_count) != 0)
	Cilk_poll();
      Cilk_fence();
      
      if (options->statlevel >= 4)
	Cilk_print_time_statistics(options->nproc);
      if (options->statlevel >= 5)
	Cilk_summarize_time_statistics(options->nproc);
      if (options->statlevel >= 6)
	Cilk_print_detailed_time_statistics(options->nproc);
    }
}
void Cilk_timing_done_machine(const Options *options)
{
  if (USE_MACHINE(Machine) == 0)
    Cilk_free_machine(USE_MACHINE(timer_array), options->nproc * sizeof(StateInfo));
}
void Cilk_timing_done_global(const Options *options)
{
}

#endif
